"""Print the command line to cd to a folder inside the project folder."""
from . import DodoCommand
import os
import sys


class Command(DodoCommand):  # noqa
    def add_arguments_imp(self, parser):  # noqa
        """
        Entry point for subclassed commands to add custom arguments.
        """
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            '--src',
            action="store_true",
            dest="cd_to_src",
            help='cd to the src directory')
        group.add_argument(
            '--build',
            action="store_true",
            dest="cd_to_build",
            help='cd to the build directory')
        group.add_argument(
            '--install',
            action="store_true",
            dest="cd_to_install",
            help='cd to the install directory')
        group.add_argument(
            '--system',
            action="store_true",
            dest="cd_to_system",
            help='cd to the dodo commands system directory')
        group.add_argument(
            '--res',
            action="store_true",
            dest="cd_to_res",
            help='cd to the res directory for this dodo commands project')

    def handle_imp(  # noqa
            self, cd_to_src, cd_to_build, cd_to_install, cd_to_system,
            cd_to_res, **kwargs):
        sys.stdout.write("cd ")
        if cd_to_src:
            src_dir = self.get_config("/ROOT/src_dir")
            print(src_dir)
        elif cd_to_build:
            build_dir = self.get_config("/ROOT/build_dir")
            print(build_dir)
        elif cd_to_install:
            project_dir = self.get_config("/ROOT/project_dir")
            print(os.path.join(project_dir, "install"))
        elif cd_to_system:
            print(self.get_config("/ROOT/system_dir"))
        elif cd_to_res:
            print(
                os.path.join(
                    self.get_config("/ROOT/project_dir"), "dodo_commands",
                    "res"))
        else:
            print(self.get_config("/ROOT/project_dir"))



# noqa
from dodo_commands.extra.standard_commands import DodoCommand

class Command(DodoCommand):  # noqa
    help = ""
    decorators = ['docker']

    def handle_imp(self, **kwargs):  # noqa
        self.runcmd(
            [
                "mkdir",
                "-p",
                self.get_config("/LESS/output_dir")
            ]
        )

        self.runcmd(
            [
                "autoless",
                ".",
                self.get_config("/LESS/output_dir")
            ],
            cwd=self.get_config("/LESS/src_dir")
        )



"""Finds a directory or file inside the current project."""
from . import DodoCommand
from dodo_commands.framework.config import (CommandPath, get_project_dir)
import os
import sys


class Command(DodoCommand):  # noqa
    def add_arguments_imp(self, parser):  # noqa
        """
        Entry point for subclassed commands to add custom arguments.
        """
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            '--src',
            action="store_true",
            dest="which_src",
            help='Print where the src directory is')
        group.add_argument(
            '--build',
            action="store_true",
            dest="which_build",
            help='Print where the build directory is')
        group.add_argument(
            '--project',
            action="store_true",
            dest="which_project",
            help='Print where the project directory is')
        group.add_argument(
            '--config',
            action="store_true",
            dest="which_config",
            help='Print where the config file is')
        group.add_argument(
            '--script',
            dest="which_script",
            help='Print where the dodo command script with given name is')
        group.add_argument(
            '--system',
            action="store_true",
            dest="which_system",
            help='Print where the dodo commands system folder is')
        group.add_argument(
            '--res',
            action="store_true",
            dest="which_res",
            help='Print where the res folder is for this dodo commands project')

    def handle_imp(  # noqa
            self, which_src, which_build, which_project, which_config,
            which_script, which_system, which_res, **kwargs):
        build_dir = self.get_config("/ROOT/build_dir", "")
        src_dir = self.get_config("/ROOT/src_dir", "")
        project_dir = self.get_config("/ROOT/project_dir", get_project_dir())

        if which_src:
            print(src_dir)
        elif which_build:
            print(build_dir)
        elif which_project:
            print(project_dir)
        elif which_config:
            print(
                os.path.join(project_dir, "dodo_commands", "res",
                             "config.yaml"))
        elif which_system:
            print(self.get_config("/ROOT/system_dir"))
        elif which_script:
            command_path = CommandPath(project_dir)
            for item in command_path.items:
                script_path = os.path.join(item.full_path,
                                           which_script + ".py")
                if os.path.exists(script_path):
                    sys.stdout.write(script_path + "\n")
        elif which_res:
            print(
                os.path.join(
                    self.get_config("/ROOT/project_dir"), "dodo_commands",
                    "res"))
        else:
            print(self.get_config("/ROOT/project_name"))



"""Run the webpack command."""
import argparse
from dodo_commands.extra.standard_commands import DodoCommand


class Command(DodoCommand):  # noqa
    decorators = ['docker']

    def add_arguments_imp(self, parser):  # noqa
        parser.add_argument(
            '--args',
            dest="webpack_args",
            required=False,
            default=[],
            nargs=argparse.REMAINDER
        )

    def handle_imp(self, webpack_args, **kwargs):  # noqa
        self.runcmd(
            ["webpack", "--watch-stdin", webpack_args],
            cwd=self.get_config("/WEBPACK/webpack_dir")
        )



"""Module for working with the dodo configurations."""

from dodo_commands.framework.command_error import CommandError
import glob
import os
import re
import sys
import yaml


def get_project_dir():
    """Return the root dir of the current project."""
    return os.environ['DODO_COMMANDS_PROJECT_DIR']


class ConfigIO:
    """Read and write the dodo configuration."""

    def __init__(self, config_base_dir=None):
        """Arg config_base_dir is where config files are searched.

        Arg config_base_dir defaults to get_project_dir()/dodo_commands/res.
        """
        self.config_base_dir = (
            os.path.join(get_project_dir(), "dodo_commands", "res")
            if config_base_dir is None
            else config_base_dir
        )

    def _path_to(self, post_fix_paths):
        """Return path composed of config_base_dir and post_fix_paths list."""
        return os.path.join(self.config_base_dir, *post_fix_paths)

    def _layers(self, config):
        result = []
        layer_dir = config.get('ROOT', {}).get('layer_dir', '')
        layers = config.get('ROOT', {}).get('layers', [])
        for layer_filename in layers:
            result.append(
                self._path_to([layer_dir, layer_filename])
            )
        return result

    def _add_layer(self, config, layer):
        for section in (layer or {}):
            if section not in config:
                config[section] = layer[section]
                continue
            for key, value in layer[section].items():
                has_key = key in config[section]
                if has_key and isinstance(value, type(dict())):
                    config[section][key].update(value)
                elif has_key and isinstance(value, type(list())):
                    config[section][key].extend(value)
                else:
                    config[section][key] = value

    def load(self, config_filename='config.yaml', load_layers=True):
        """Get configuration."""
        full_config_filename = self._path_to([config_filename])
        if not os.path.exists(full_config_filename):
            return None

        with open(full_config_filename) as f:
            config = yaml.load(f.read())
        if load_layers:
            for layer_filename in self._layers(config):
                if os.path.exists(layer_filename):
                    self._add_layer(config, self.load(layer_filename))
                else:
                    print("Warning: layer not found: %s" % layer_filename)

        return config

    def save(self, config, config_filename='config.yaml'):
        """Write config to config_filename."""
        content = yaml.dump(config, default_flow_style=False, indent=4)
        with open(self._path_to([config_filename]), 'w') as f:
            formatted_content = re.sub(
                r'^([0-9_A-Z]+\:)$',
                r'\n\1',
                content,
                flags=re.MULTILINE
            )
            return f.write(
                # swallow the leading newline
                formatted_content[1:]
                if formatted_content[0] == '\n' else
                formatted_content
            )


def load_dodo_config():
    """Load the project's dodo config and expand it."""
    def _add_to_config(config, section, key, value):
        if section in config:
            if key not in config[section]:
                config[section][key] = value

    config = ConfigIO().load() or dict(ROOT={})
    project_dir = get_project_dir()
    _add_to_config(
        config, "ROOT", "project_name", os.path.basename(project_dir))
    _add_to_config(
        config, "ROOT", "project_dir", project_dir)
    ConfigExpander().run(config)
    return config


class KeyNotFound(CommandError):
    """Raised if an xpath is not found in a configuration."""

    pass


class Key:
    """Access a nested value in a dict."""

    def __init__(self, config, xpath):  # noqa
        self.config = config
        # the list of sub-keys along the path to an item in self.config
        self.xpath = xpath

    def child(self, subkey):
        """Return key for child with subkey."""
        return Key(self.config, list(self.xpath) + [subkey])

    def _step_into(self, node, subkey):
        try:
            if isinstance(node, type(dict())):
                return node[subkey]
            if isinstance(node, type(list())):
                return node[int(subkey)]
        except:
            raise KeyNotFound("Cannot get value at %s\n" % self)

    def get(self):  # noqa
        node = self.config
        for subkey in self.xpath:
            node = self._step_into(node, subkey)
        return node

    def set(self, value):  # noqa
        node = self.config
        for subkey in self.xpath[:-1]:
            node = self._step_into(node, subkey)
        node[self.xpath[-1]] = value

    def remove(self):  # noqa
        node = self.config
        for subkey in self.xpath[:-1]:
            node = self._step_into(node, subkey)
        del node[self.xpath[-1]]

    def __repr__(self):  # noqa
        return str(self.xpath)


class ConfigExpander:
    """Expand environment variables and references in the config."""

    _regexp = r"\$\{(/[^\}]*)\}"

    @classmethod
    def _is_fully_expanded(cls, key, processed_xpaths):
        if key.xpath in processed_xpaths:
            return True

        value = key.get()
        if cls._is_leaf(value):
            return False

        return all([
            cls._is_fully_expanded(key.child(x), processed_xpaths)
            for x in cls._subkeys_of(value)
        ])

    @classmethod
    def _must_evaluate(cls, key):
        last_subkey = key.xpath[-1]
        return (
            isinstance(last_subkey, type(str())) and
            last_subkey.endswith("_EVAL")
        )

    @classmethod
    def _evaluate(cls, key):
        value = key.get()
        if cls._is_leaf(value):
            try:
                key.set(eval(value))
            except:
                raise CommandError(
                    "Cannot evaluate value '%s' for key %s\n" % (value, key)
                )
        else:
            for k in cls._subkeys_of(value):
                if cls._is_leaf(value[k]):
                    cls._evaluate(key.child(k))

    @classmethod
    def _is_leaf(cls, value):
        return not(
            isinstance(value, type(dict())) or
            isinstance(value, type(list()))
        )

    @classmethod
    def _subkeys_of(cls, value):
        if isinstance(value, type(dict())):
            return [k for k in value]
        if isinstance(value, type(list())):
            return range(len(value))
        return []

    @classmethod
    def _key_for_match(cls, matched_str, config):
        key = Key(config, [])
        for index in [ix for ix in matched_str.split("/") if ix]:
            if isinstance(key.get(), type(list())):
                subkey = int(index)
            else:
                subkey = index
            key = key.child(subkey)
        return key

    @classmethod
    def _expand(cls, value, config, processed_xpaths):
        matches = re.finditer(cls._regexp, value)
        for match in reversed([m for m in matches]):
            replacement = cls._key_for_match(match.group(1), config)
            if replacement.xpath not in processed_xpaths:
                return False, None
            value = (
                value[:match.start()] +
                str(replacement.get()) +
                value[match.end():]
            )
        return True, os.path.expandvars(value)

    @classmethod
    def _expand_value_for(cls, key, processed_xpaths):
        value = key.get()
        if not cls._is_leaf(value):
            return cls._is_fully_expanded(key, processed_xpaths), None
        if not isinstance(value, type(str())):
            return True, None
        return cls._expand(
            value, key.config, processed_xpaths
        )

    @classmethod
    def _for_each_xpath(cls, config, node, f, xpath=None):
        if xpath is None:
            xpath = []

        for subkey in cls._subkeys_of(node):
            value = node[subkey]
            f(config, list(xpath) + [subkey])
            cls._for_each_xpath(
                config, value, f, list(xpath) + [subkey]
            )

    @classmethod
    def _expanded_key(cls, key, processed_xpaths):
        """Return key with variable expansion on last subkey.

        Return None if not all variables could be expanded.
        """
        last_subkey = key.xpath[-1]
        if not isinstance(last_subkey, type(str())):
            return key

        is_expanded, expanded_last_subkey = cls._expand(
            last_subkey, key.config, processed_xpaths
        )
        if not is_expanded:
            return None
        return Key(
            key.config, key.xpath[:-1] + [expanded_last_subkey]
        )

    @classmethod
    def run(cls, config):  # noqa
        all_keys = []
        cls._for_each_xpath(
            config, config,
            lambda config, xpath: all_keys.append(Key(config, xpath))
        )

        processed_xpaths = []
        while all_keys:
            for key in list(all_keys):
                expanded_key = cls._expanded_key(key, processed_xpaths)
                if expanded_key is None:
                    continue

                is_expanded, expanded_value = cls._expand_value_for(
                    key, processed_xpaths
                )

                if is_expanded:
                    if expanded_value is not None:
                        expanded_key.set(expanded_value)
                        if expanded_key.xpath != key.xpath:
                            key.remove()
                    if cls._must_evaluate(expanded_key):
                        cls._evaluate(expanded_key)
                    processed_xpaths.append(expanded_key.xpath)
                    all_keys.remove(key)


class CommandPath:
    """Read search paths for command scripts from the configuration."""

    class Item:  # noqa
        sys_path = ""
        module_path = ""

        @property
        def full_path(self):  # noqa
            return os.path.join(self.sys_path, self.module_path)

    def __init__(self, project_dir, config_base_dir=None):
        """config_base_dir is the directory where config files are searched.

        The config_base_dir arg defaults to project_dir/dodo_commands/res.
        """
        if config_base_dir is None:
            config_base_dir = os.path.join(project_dir, "dodo_commands", "res")

        config = load_dodo_config()
        excluded_dirs = [
            x.full_path for x in self._collect_items(
                config.get('ROOT', {}).get('command_path_exclude', []),
            )
        ]
        self.items = [
            x for x in self._collect_items(
                config.get('ROOT', {}).get('command_path', [])
            )
            if x.full_path not in excluded_dirs
        ]

    def _collect_items(self, patterns):
        items = []
        for pattern in patterns:
            try:
                prefix, postfix = pattern
                prefix = os.path.expanduser(prefix)
            except:
                raise CommandError(
                    "Items in command_path should be of " +
                    "type [<sys-path>, <module-path>]."
                )

            for command_dir in [
                x for x in glob.glob(os.path.join(prefix, postfix))
                if os.path.isdir(x)
            ]:
                item = CommandPath.Item()
                item.sys_path = prefix
                item.module_path = os.path.relpath(command_dir, prefix)
                items.append(item)
        return items

    def extend_sys_path(self):  # noqa
        for item in self.items:
            if item.sys_path not in sys.path:
                sys.path.append(item.sys_path)



# noqa
from dodo_commands.extra.standard_commands import DodoCommand


class Command(DodoCommand):  # noqa
    help = ""
    decorators = ["docker"]

    def handle_imp(self, **kwargs):  # noqa
        self.runcmd(
            [
                self.get_config("/PYTEST/pytest"),
                "-k", "test_accessibility_password_reset_recover_page"
            ],
            cwd=self.get_config("/PYTEST/src_dir")
        )



from setuptools import setup
import os
import sys


# in setup.py
if 'bdist_wheel' in sys.argv:
    raise RuntimeError("This setup.py does not support wheels")

setup(name='dodo_commands',
      version='0.3.4',
      description='Project-aware development environments, inspired by django-manage',
      url='https://github.com/mnieber/dodo_commands',
      download_url='https://github.com/mnieber/dodo_commands/tarball/0.3.4',
      author='Maarten Nieber',
      author_email='hallomaarten@yahoo.com',
      license='MIT',
      packages=[
          'dodo_commands',
          'dodo_commands.framework',
      ],
      package_data={
          'dodo_commands': [
              'extra/__init__.py',
              'extra/git_commands/*.py',
              'extra/git_commands/*.meta',
              'extra/standard_commands/*.py',
              'extra/standard_commands/*.meta',
              'extra/standard_commands/decorators/*.py',
              'extra/webdev_commands/*.py',
              'extra/webdev_commands/*.meta',
          ]
      },
      data_files=[
          (os.path.expanduser('~/.dodo_commands'), ['res/config']),
          (os.path.expanduser('~/.dodo_commands/default_commands'), ['res/default_commands/__init__.py']),
      ],
      entry_points={
          'console_scripts': [
              'dodo-activate=dodo_commands.dodo_activate:main',
              'dodo-install-default-commands=dodo_commands.dodo_install_commands:main',
          ]
      },
      install_requires=[
          'argcomplete',
          'plumbum',
          'PyYaml',
          'six',
      ],
      zip_safe=False)
